/*
 * tscat
 *
 * a filter that reads from stdin, prepends a timestamp to every line,
 * and writes to stdout
 *
 * $Id: tscat.c 10 2006-02-20 00:07:47Z greg $
 *
 * Greg Ward <gward-tscat at python dot net> 2005/07/30
 *
 * build instructions:
 *   make tscat
 *
 * installation instructions:
 *   make install
 */

/*
Copyright (c) 2006 Gregory P. Ward.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>

#define GNU_SOURCE
#include <getopt.h>

void
usage(char* error)
{
    fprintf(stdout,
            "usage: tscat [options]\n"
            "\n"
            "tscat simply reads from stdin and writes to stdout"
            "\n"
            "options:\n"
            "  -r, --relative   show relative times (elapsed since start of run)\n"
            "  -d, --delta      show deltas (elapsed since last line)\n"
            "\n");
            
    if (error != NULL)
        fprintf(stderr, "error: %s\n", error);

    exit(2);
}

typedef enum {
    ABSOLUTE_TIME,
    RELATIVE,
    DELTA,
} displaymode_t;

void
parse_args(int argc, char** argv, displaymode_t* mode)
{
    struct option options[] = {
        { "relative", 0, NULL, (int) 'r' },
        { "delta",    0, NULL, (int) 'd' },
        { 0, 0, 0, 0 },
    };

    int opt;
    while ( (opt = getopt_long(argc, argv, "rd", options, NULL)) != -1 ) {
        switch (opt) {
            case 'r':
                *mode = RELATIVE;
                break;
            case 'd':
                *mode = DELTA;
                break;
            case '?':
                usage(NULL);
        }
    }

    if (optind < argc)
        usage("too many arguments");
}


void
ftime(char* buf, size_t size, struct timeval *tv)
{
    buf[0] = '[';
    struct tm* tm = localtime(&tv->tv_sec);
    strftime(buf+1, size-1, "%H:%M:%S", tm);
    snprintf(buf+9, size-9, ".%06d]", (int) tv->tv_usec);
}

void
fputtime(struct timeval *tv, FILE* file)
{
    char buf[40];
    ftime(buf, 40, tv);
    fputs(buf, file);
}

void
fputdelta(struct timeval *tv1, struct timeval *tv2, FILE* file)
{
    /* compute the delta as a (sec, usec) pair */
    int delta_sec = tv2->tv_sec - tv1->tv_sec;
    int delta_usec = tv2->tv_usec - tv1->tv_usec;
    if (delta_usec < 0) {
        delta_sec--;
        delta_usec += 1000000;
    }

#if 0
    /* print the original times (for debugging) */
    char buf[40];
    ftime(buf, 40, tv1);
    fputs(buf, file);
    fputs(" .. ", file);
    ftime(buf, 40, tv2);
    fputs(buf, file);
    fputs(" = ", file);
#endif

    /* and print the delta */
    fprintf(file, "[%3d.%06d]", delta_sec, delta_usec);
}


int
main(int argc, char** argv)
{
#define BUFSIZE 1024
    char buf[BUFSIZE];

    displaymode_t mode = ABSOLUTE_TIME;
    parse_args(argc, argv, &mode);

    struct timeval start_tv;
    struct timeval previous_tv;
    struct timeval now_tv;

    /* initialize start and previous times -- used by RELATIVE and DELTA
       modes */
    gettimeofday(&start_tv, NULL);
    previous_tv = start_tv;

#if 0
    fputs("start_tv = ", stdout);
    fputtime(&start_tv, stdout);
    fputs("\nprevious_tv = ", stdout);
    fputtime(&previous_tv, stdout);
    putc('\n', stdout);
#endif

    int need_ts = 1;    /* no timestamp written yet for current line */
    while (fgets(buf, BUFSIZE, stdin) != NULL) {
        if (need_ts) {
            gettimeofday(&now_tv, NULL);
            switch (mode) {
                case ABSOLUTE_TIME:
                    fputtime(&now_tv, stdout);
                    break;
                case RELATIVE:
                    fputdelta(&start_tv, &now_tv, stdout);
                    break;
                case DELTA:
                    fputdelta(&previous_tv, &now_tv, stdout);
                    break;
            }
            putc(' ', stdout);
            previous_tv = now_tv;
        }

        fputs(buf, stdout);
        need_ts = (buf[strlen(buf)-1] == '\n');
    }
    return 0;
}
